@@ -3,12 +3,26 @@
 import logging
 from abc import ABC, abstractmethod
 from contextlib import contextmanager
+from dataclasses import dataclass
 from typing import List, Tuple
 from urllib.parse import parse_qs, urlparse
 
 from dinao.backend.errors import ConfigurationError
 
 
+@dataclass
+class ColumnDescriptor:
+    """Describes a column in a result set."""
+
+    name: str
+    type_code: int
+    display_size: int = None
+    internal_size: int = None
+    precision: int = None
+    scale: int = None
+    null_ok: bool = None
+
+
 class ResultSet:
     """Basic interface definition for result sets (a.k.a rows) returned from Database queries."""
 
@@ -18,31 +32,44 @@ class ResultSet:
         :param cursor: the underlying DB API 2.0 cursor being wrapped by this object.
         """
         self._cursor = cursor
+        self._columns = None
+        self._description = None
 
-    def fetchone(self) -> tuple:
+    def fetchone(self) -> Tuple:
         """Fetch one result tuple from the underlying cursor.
 
         If no results are left, None is returned.
 
-        :return: a tuple representing a result row or None
+        :returns: a tuple representing a result row or None
         """
         return self._cursor.fetchone()
 
-    def fetchall(self) -> List[tuple]:
+    def fetchall(self) -> List[Tuple]:
         """Fetch the *remaining* result tuples from the underlying cursor.
 
         If no results are left, an empty list is returned.
 
-        :return: a list of tuples that are the remaining results of the underlying cursor.
+        :returns: a list of tuples that are the remaining results of the underlying cursor.
         """
         return self._cursor.fetchall()
 
-    def columns(self) -> Tuple[str]:
-        """Return a tuple representing the column names of the result set.
+    @property
+    def description(self) -> Tuple[ColumnDescriptor]:
+        """Return a sequence of column descriptions representing the result set.
 
-        :return: a tuple of strings that are column names matching the order of results.
+        :returns: a tuple of ColumnDescriptors
         """
-        return tuple([str(d[0]) for d in self._cursor.description])
+        if not self._description:
+            self._description = tuple([ColumnDescriptor(*d) for d in self._cursor.description])
+        return self._description
+
+    @property
+    def rowcount(self) -> int:
+        """Return the row count of the result set.
+
+        :returns: the integer count of the rows in the result set
+        """
+        return self._cursor.rowcount
 
 
 class Connection(ABC):
@@ -58,6 +85,15 @@ class Connection(ABC):
         self._cnx = cnx
         self._auto_commit = auto_commit
 
+    @property
+    def autocommit(self):
+        """Whether or not commit is called after every call to query(...) and execute(...)."""
+        return self._auto_commit
+
+    @autocommit.setter
+    def autocommit(self, value: bool):
+        self._auto_commit = value
+
     def commit(self):
         """Commit changes for this connection / transaction to the database."""
         self._cnx.commit()
@@ -80,8 +116,10 @@ class Connection(ABC):
         """
         cursor = self._cnx.cursor()
         self._execute(cursor, sql, params)
-        yield ResultSet(cursor)
-        cursor.close()
+        try:
+            yield ResultSet(cursor)
+        finally:
+            cursor.close()
 
     def execute(self, sql: str, params: tuple = None, commit: bool = None) -> int:
         """Execute the given SQL as a statement with the given parameters and return the affected row count.
@@ -108,7 +146,7 @@ class ConnectionPool(ABC):
 
         The db_url is expected to be in the following format::
 
-            "{db_backend}+{driver}://{username}:{password}@{hostname}:{port}/{db_name}?{optional_ars}"
+            "{dialect}+{driver}://{username}:{password}@{hostname}:{port}/{db_name}?{optional_args}"
 
         :param db_url: a url with the described format
         """
@@ -116,12 +154,11 @@ class ConnectionPool(ABC):
         self._raw_db_url = db_url
         self._db_url = urlparse(self._raw_db_url)
         self._args = parse_qs(self._db_url.query)
-        self._defer_pool = self._get_arg("defer", bool, False)
 
     @staticmethod
     def _strict_bool(value: str):
         if value.lower() not in ["true", "false"]:
-            raise ValueError()
+            raise ValueError(f"Cannot cast '{value}' to bool")
         return value.lower() == "true"
 
     def _raise_for_unexpected_args(self):
@@ -131,7 +168,7 @@ class ConnectionPool(ABC):
 
     def _get_arg(self, name: str, expected_type, default=None):
         if name not in self._args:
-            self.logger.debug(f'No "{name}" specified, defaulting to {default}')
+            self.logger.debug(f"No '{name}' specified, defaulting to {default}")
             return default
         caster = expected_type if expected_type is not bool else self._strict_bool
         try:
@@ -139,10 +176,10 @@ class ConnectionPool(ABC):
                 assert len(self._args.get(name)) == 1
                 return caster(self._args.pop(name)[0])
             return self._args.pop(name)
-        except AssertionError:
-            raise ConfigurationError(f'Invalid argument "{name}": only a single value must be specified')
-        except ValueError:
-            raise ConfigurationError(f'Invalid argument "{name}": must be {expected_type.__name__}')
+        except AssertionError as x:
+            raise ConfigurationError(f"Invalid argument '{name}': only a single value must be specified") from x
+        except ValueError as x:
+            raise ConfigurationError(f"Invalid argument '{name}': must be {expected_type.__name__}") from x
 
     @property
     @abstractmethod
@@ -159,3 +196,8 @@ class ConnectionPool(ABC):
     def release(self, cnx: Connection):
         """Release a connection back to the underlying pool."""
         pass  # pragma: no cover
+
+    @abstractmethod
+    def dispose(self):
+        """Close the pool and clean up any resources it was using."""
+        pass  # pragma: no cover
